
    class Tile {
        constructor({ id, name, hostname, os_name, os_version, os_type, main_ip, ramUsage, totalMemory, usedMemory, cpuUsage, cpuCoreCount, cpuClockSpeed, color, apiUrl, updateInterval = 5000, type = 'generic', icon = '', data = '', onClick = null }) {
            this.id = id;
            this.name = name;
            this.hostname = hostname;
            this.os_name = os_name;
            this.os_version = os_version;
            this.os_type = os_type;
            this.main_ip = main_ip;
            this.ramUsage = ramUsage;
            this.totalMemory = totalMemory;
            this.usedMemory = usedMemory;
            this.cpuUsage = cpuUsage;
            this.cpuCoreCount = cpuCoreCount;
            this.cpuClockSpeed = cpuClockSpeed;
            this.color = color;
            this.apiUrl = apiUrl;
            this.updateInterval = updateInterval;
            this.type = type;
            this.icon = icon;
            this.onClick = onClick;
            this.element = null;
            this.data = data;
            this.windowInstance = null;

            // Store historical CPU and RAM data
            this.cpuHistory = [];
            this.ramHistory = [];
            this.datetimeHistory = [];
            this.isActive = true;

            this.createTile();
            this.updateTileContent();

            setInterval(() => this.updateTileContent(), this.updateInterval);
        }

        createTile() {
            const tileContainer = document.querySelector('.tile-container');
            this.element = document.createElement('div');
            this.element.className = 'tile flex-container';

            if (this.color && this.color !== masterTileColor) {
                this.element.setAttribute('data-custom-color', 'true');
            } else {
                this.element.setAttribute('data-custom-color', 'false');
            }

            this.element.style.backgroundColor = this.isActive ? (this.color || masterTileColor) : 'rgba(128,128,128,0.5)'; // Greyed out if inactive
            this.element.setAttribute('draggable', 'true');

            const headerWrapper = document.createElement('div');
            headerWrapper.className = 'tile-header-wrapper';

            if (this.icon) {
                const iconElement = document.createElement('img');
                iconElement.src = this.icon;
                iconElement.className = 'tile-icon';
                headerWrapper.appendChild(iconElement);
            }

            const textElement = document.createElement('div');
            textElement.className = 'tile-text';
            textElement.textContent = this.name;
            headerWrapper.appendChild(textElement);

            const dataElement = document.createElement('div');
            dataElement.className = 'tile-data';
            dataElement.innerHTML = this.data;
            headerWrapper.appendChild(dataElement);

            const editButton = document.createElement('button');
            editButton.className = 'tile-edit-button';
            editButton.innerHTML = '&#9881;';
            editButton.onclick = (e) => {
                e.stopPropagation();
                this.openSettingsWindow();
            };
            this.element.appendChild(editButton);

            this.element.appendChild(headerWrapper);

            const contentElement = document.createElement('div');
            contentElement.className = 'tile-content';
            this.element.appendChild(contentElement);
            tileContainer.appendChild(this.element);

            if (this.onClick) {
                this.element.addEventListener('click', () => this.onClick(this));
            }

            this.addDragAndDropHandlers();
        }

        addDragAndDropHandlers() {
            this.element.addEventListener('dragstart', (e) => {
                e.dataTransfer.setData('text/plain', this.element.dataset.index);
                this.element.classList.add('dragging');
            });
    
            this.element.addEventListener('dragend', () => {
                this.element.classList.remove('dragging');
            });
    
            this.element.addEventListener('dragover', (e) => {
                e.preventDefault();
                this.element.classList.add('drag-over');
            });
    
            this.element.addEventListener('dragleave', () => {
                this.element.classList.remove('drag-over');
            });
    
            this.element.addEventListener('drop', (e) => {
                e.preventDefault();
                this.element.classList.remove('drag-over');
    
                const tileContainer = document.querySelector('.tile-container');
                const draggedTileIndex = e.dataTransfer.getData('text/plain');
                const draggedTile = tileContainer.querySelector(`[data-index="${draggedTileIndex}"]`);
    
                if (draggedTile !== this.element) {
                    tileContainer.insertBefore(draggedTile, this.element.nextSibling);
                }
            });
        }

        openSettingsWindow() {
            const settingsWindow = new Window({
                title: 'Edit Tile Settings',
                width: 300,
                height: 250,
                content: `
                    <label for="tile-title">Title:</label>
                    <input type="text" id="tile-title" value="${this.name}" aria-label="Edit Tile Title">
                    <br />
                    <label for="tile-color">Color:</label>
                    <br />
                    <input type="color" id="tile-color" value="${this.color || masterTileColor}" aria-label="Choose Tile Color">
                `
            });

            const titleInput = settingsWindow.windowElement.querySelector('#tile-title');
            titleInput.addEventListener('input', (e) => {
                this.name = e.target.value;
                this.element.querySelector('.tile-text').textContent = e.target.value;
            });
    
            const colorInput = settingsWindow.windowElement.querySelector('#tile-color');
            colorInput.addEventListener('input', (e) => {
                this.color = e.target.value;
                this.element.style.backgroundColor = e.target.value;
                this.element.setAttribute('data-custom-color', 'true');
            });
        }


        openHostDetailsWindow() {
            // If window already exists, just bring it to the front
            if (this.windowInstance) {
                if (this.chartInstance) {
                    this.chartInstance.destroy();  // Destroy chart when window is closed
                }
                this.windowInstance.closeWindow();
                this.windowInstance = null;
                return;
            }
        
            const windowContent = `
                <strong>Memory:</strong> ${(Math.floor(this.totalMemory * 100) / 100).toFixed(2)} GB &nbsp;&nbsp; 
                <strong>CPU:</strong> ${(Math.floor(this.cpuClockSpeed * 100) / 100).toFixed(2)} Ghz (x${this.cpuCoreCount})
                <canvas id="combinedChart-${this.hostname}" width="350" height="275" class="data-graph"></canvas>
            `;
        
            // Create a new window
            this.windowInstance = new Window({
                title: `Host Details - ${this.hostname}`,
                width: 350,
                height: 350,
                content: windowContent
            });
        
            // Generate charts for the host
            this.generateCharts();
        
            // Clean up the window instance on close
            this.windowInstance.windowElement.querySelector('.window-close').addEventListener('click', () => {
                if (this.chartInstance) {
                    this.chartInstance.destroy();  // Destroy chart when window is closed
                }
                this.windowInstance = null;
            });
        }        
        
        generateCharts() {
            const canvasId = `combinedChart-${this.hostname}`; // Unique canvas ID for each tile
            const ctx = document.getElementById(canvasId).getContext('2d');
        
            // Destroy previous chart instance if it exists
            if (this.chartInstance) {
                this.chartInstance.destroy();
            }
        
            // Algorithm to control the number of points displayed on the chart
            const maxEntries = 10; // Maximum number of entries to display
            const totalEntries = this.datetimeHistory.length;
            let stepSize = 1;
        
            if (totalEntries > maxEntries) {
                stepSize = Math.ceil(totalEntries / maxEntries); // Reduce number of points shown
            }
        
            // Grab every X-th entry based on the stepSize
            const filteredLabels = this.datetimeHistory
                .filter((_, index) => index % stepSize === 0)
                .map(date => new Date(date).toLocaleTimeString());
        
            const filteredCpuHistory = this.cpuHistory.filter((_, index) => index % stepSize === 0);
            const filteredRamHistory = this.ramHistory.filter((_, index) => index % stepSize === 0);
        
            // Create the combined chart and store the instance
            this.chartInstance = new Chart(ctx, {
                type: 'line',
                data: {
                    labels: filteredLabels,
                    datasets: [
                        {
                            label: 'CPU Usage (%)',
                            data: filteredCpuHistory,
                            borderColor: 'rgba(75, 192, 192, 1)',
                            borderWidth: 2,
                            fill: false
                        },
                        {
                            label: 'RAM Usage (%)',
                            data: filteredRamHistory,
                            borderColor: 'rgba(153, 102, 255, 1)',
                            borderWidth: 2,
                            fill: false
                        }
                    ]
                },
                options: {
                    scales: {
                        x: {
                            display: true,
                            title: {
                                display: false,
                                text: 'Time'
                            }
                        },
                        y: {
                            beginAtZero: true,
                            max: 100
                        }
                    }
                }
            });
        }
        
        updateTileContent(newData) {
            if (newData) {
                const currentTime = new Date().getTime();
                this.ramUsage = newData.memoryUsage;
                this.cpuUsage = newData.cpuUsage;
                this.main_ip = newData.ip_address;
                this.cpuClockSpeed = newData.cpuClockSpeed;
                this.cpuCoreCount = newData.cpuCoreCount;
        
                // Update history arrays
                this.cpuHistory.push(this.cpuUsage);
                this.ramHistory.push(this.ramUsage);
                this.datetimeHistory.push(newData.datetime);
        
                // Limit history arrays to 300 entries
                if (this.cpuHistory.length > 300) { 
                    this.cpuHistory.shift();
                    this.ramHistory.shift();
                    this.datetimeHistory.shift();
                }
        
                // Only check the last 30 timestamps if we have at least 30 entries
                const lastTimestamps = this.datetimeHistory.slice(-30);
                const isSameTimestamp = lastTimestamps.length === 30 && lastTimestamps.every(
                    (ts) => ts === newData.datetime
                );
        
                // Consider the host active if fewer than 30 timestamps exist, or if timestamps differ
                this.isActive = lastTimestamps.length < 30 || !isSameTimestamp;
        
                // Set the tile color and opacity based on activity status
                if (this.isActive) {
                    this.element.setAttribute('data-offline-color', 'false');
                    this.element.style.opacity = storedTileOpacity; // Set to default opacity when active
                    const hasCustomColor = this.element.getAttribute('data-custom-color') === 'true';
                    if (hasCustomColor) {
                        this.element.style.backgroundColor = this.color;
                    }
                } else {
                    this.element.setAttribute('data-offline-color', 'true');
                    this.element.style.backgroundColor = 'rgba(128,128,128,0.5)';
                    this.element.style.opacity = '0.5'; // Set lower opacity when inactive (adjust as needed)
                }
        
                this.updateTileDisplay();
            }
        }                             

        updateTileDisplay() {
            const isListView = this.element.classList.contains('list-view');
        
            if (isListView) {
                const cpuUsageBar = `
                    <div class="vertical-bar-container">
                        <div class="vertical-bar cpu" style="height: ${this.cpuUsage}%;"></div>
                        <div class="labelvalue">${this.cpuUsage}%</div>
                        <div class="label">C</div>
                    </div>
                `;
                const ramUsageBar = `
                    <div class="vertical-bar-container">
                        <div class="vertical-bar ram" style="height: ${this.ramUsage}%;"></div>
                        <div class="labelvalue">${this.ramUsage}%</div>
                        <div class="label">R</div>
                    </div>
                `;
        
                this.data = `
                    <table>
                        <tr>
                            <td class='sys-info-container'>
                                OS: ${this.os_name}<br />
                                Version: ${this.os_version}<br />
                                IP Address: ${this.main_ip}
                            </td>
                            <td>
                                <div class="tile-graph">
                                    ${cpuUsageBar}
                                    ${ramUsageBar}
                                </div>
                            </td>
                        </tr>
                    </table>
                `;
            } else {
                const cpuUsageBar = this.createStatusBar(this.cpuUsage, 'cpu-usage-bar');
                const ramUsageBar = this.createStatusBar(this.ramUsage, 'ram-usage-bar');
        
                this.data = `
                    OS: ${this.os_name}<br />
                    Version: ${this.os_version}<br />
                    IP Address: ${this.main_ip}<br />
                    <div class="tile-graph">
                        <span class="status-bar-label">CPU:</span> ${cpuUsageBar}
                    </div>
                    <div class="tile-graph">
                        <span class="status-bar-label">RAM:</span> ${ramUsageBar}
                    </div>
                `;
            }
        
            this.element.querySelector('.tile-data').innerHTML = this.data;
        }
        
        createStatusBar(percentage, className) {
            return `
                <div class="status-bar-container">
                    <div class="status-bar ${className}" style="width: ${percentage}%;"></div>
                    <span class="labelvalue">${percentage}%</span>
                </div>
            `;
        }
        
    }

    class Toast {
        constructor({ message, backgroundColor, autoDismiss = true, action = null, actionLabel = 'Launch', timeout = 5000 }) {
            this.message = message;
            this.backgroundColor = backgroundColor || '';
            this.autoDismiss = autoDismiss;
            this.action = action;
            this.actionLabel = actionLabel;
            this.timeout = timeout;
            this.toastElement = null;

            this.createToast();
            if (this.autoDismiss && !this.action) {
                setTimeout(() => this.dismissToast(), this.timeout);
            }
        }

        createToast() {
            let toastContainer = document.querySelector('.toast-container');
            if (!toastContainer) {
                toastContainer = document.createElement('div');
                toastContainer.className = 'toast-container';
                document.body.appendChild(toastContainer);
            }

            this.toastElement = document.createElement('div');
            this.toastElement.className = 'toast flex-container';
            this.toastElement.style.backgroundColor = this.backgroundColor;

            const messageElement = document.createElement('div');
            messageElement.className = 'toast-message';
            messageElement.textContent = this.message;
            this.toastElement.appendChild(messageElement);

            const buttonsContainer = document.createElement('div');
            buttonsContainer.className = 'toast-buttons';

            if (this.action) {
                const actionButton = document.createElement('button');
                actionButton.className = 'toast-button';
                actionButton.textContent = this.actionLabel;
                actionButton.onclick = () => {
                    this.action(this);
                    this.dismissToast();
                };
                buttonsContainer.appendChild(actionButton);
            }

            const dismissButton = document.createElement('button');
            dismissButton.className = 'toast-button';
            dismissButton.textContent = 'Dismiss';
            dismissButton.onclick = () => this.dismissToast();
            buttonsContainer.appendChild(dismissButton);

            this.toastElement.appendChild(buttonsContainer);
            toastContainer.appendChild(this.toastElement);
        }

        dismissToast() {
            if (this.toastElement) {
                this.toastElement.classList.add('hide');
                setTimeout(() => {
                    this.toastElement.remove();
                }, 300);
            }
        }
    }

    class Window {
        static activeWindows = [];

        constructor({ title = "Window", width = 350, height = 300, apiUrl = null, content = "" }) {
            this.title = title;
            this.width = width;
            this.height = height;
            this.apiUrl = apiUrl;
            this.content = content;
            this.windowElement = null;
            this.isDragging = false;
            this.offset = { x: 0, y: 0 };

            this.createWindow();
            this.setActiveWindow();
            Window.activeWindows.push(this);
            this.reIndexWindows();
        }

        createWindow() {
            this.windowElement = document.createElement('div');
            this.windowElement.className = 'window';
            
            // Set initial properties like width and height if known beforehand
            this.windowElement.style.setProperty('--window-width', `${this.width}px`);
            this.windowElement.style.setProperty('--window-height', `${this.height}px`);
        
            const header = document.createElement('div');
            header.className = 'window-header flex-container';
            header.textContent = this.title;
        
            const closeButton = document.createElement('button');
            closeButton.className = 'window-close';
            closeButton.innerHTML = '&times;';
            closeButton.onclick = () => this.closeWindow();
        
            header.appendChild(closeButton);
        
            const contentArea = document.createElement('div');
            contentArea.className = 'window-content';
            contentArea.innerHTML = this.content;
            
            this.windowElement.appendChild(header);
            this.windowElement.appendChild(contentArea);
            document.body.appendChild(this.windowElement);
        
            // Now that the window is part of the DOM, calculate the actual width and height
            const windowRect = this.windowElement.getBoundingClientRect();
            const windowWidth = windowRect.width;
            const windowHeight = windowRect.height;
        
            const screenWidth = window.innerWidth;
            const screenHeight = window.innerHeight;
            
            // Check if the device is mobile based on width and height
            const isMobile = (screenWidth <= 768) || (screenHeight <= 1024); // Mobile threshold for both orientations
            
            let centerX, centerY;
            
            if (isMobile) {
                // Center for mobile devices in both orientations
                centerX = (screenWidth / 2); // Center horizontally
                centerY = ((screenHeight / 2) - 100); // Center vertically
            } else {
                // Normal centering for larger screens
                centerX = (screenWidth - windowWidth) / 2;
                centerY = (screenHeight - windowHeight) / 2;
            
                const windowCount = Window.activeWindows.length;
                const baseOffset = 10;
                const maxOffset = 200;
            
                let offsetX = baseOffset * windowCount;
                let offsetY = baseOffset * windowCount;
            
                if (offsetX > maxOffset) offsetX = 0;
                if (offsetY > maxOffset) offsetY = 0;
            
                centerX += offsetX;
                centerY += offsetY;
            }
            
            // Set the window's position
            this.windowElement.style.left = `${centerX}px`;
            this.windowElement.style.top = `${centerY}px`;
        
            if (this.apiUrl) {
                this.loadContentFromApi();
            }
        
            this.addDragEventListeners(header);
        }                

        setActiveWindow() {
            Window.activeWindows.forEach(window => {
                window.windowElement.classList.add('inactive-window');
            });

            this.windowElement.classList.remove('inactive-window');
            this.windowElement.style.zIndex = Window.activeWindows.length + 1000;

            Window.activeWindows = Window.activeWindows.filter(win => win !== this);
            Window.activeWindows.push(this);

            Window.activeWindows.forEach((window, index) => {
                window.windowElement.style.zIndex = 1000 + index;
            });
        }

        closeWindow() {
            if (this.windowElement) {
                this.windowElement.remove();

                const isActiveWindow = Window.activeWindows[Window.activeWindows.length - 1] === this;

                Window.activeWindows = Window.activeWindows.filter(win => win !== this);

                if (isActiveWindow && Window.activeWindows.length > 0) {
                    const nextActiveWindow = Window.activeWindows[Window.activeWindows.length - 1];
                    nextActiveWindow.setActiveWindow();
                }

                this.reIndexWindows();
            }
        }

        reIndexWindows() {
            Window.activeWindows.forEach((window, index) => {
                window.windowElement.style.zIndex = 1000 + index;
            });
        }

        addDragEventListeners(header) {
            header.addEventListener('mousedown', (e) => this.onDragStart(e));
            document.addEventListener('mousemove', (e) => this.onDrag(e));
            document.addEventListener('mouseup', () => this.onDragEnd());
        }

        onDragStart(e) {
            this.isDragging = true;
            this.offset.x = e.clientX - this.windowElement.offsetLeft;
            this.offset.y = e.clientY - this.windowElement.offsetTop;
            this.setActiveWindow();
            e.preventDefault();
        }

        onDrag(e) {
            if (this.isDragging) {
                this.windowElement.style.left = `${e.clientX - this.offset.x}px`;
                this.windowElement.style.top = `${e.clientY - this.offset.y}px`;
            }
        }

        onDragEnd() {
            this.isDragging = false;
        }

        loadContentFromApi() {
            fetch(this.apiUrl)
                .then(response => response.text())
                .then(data => {
                    this.windowElement.querySelector('.window-content').innerHTML = data;
                })
                .catch(error => console.error('Error loading content:', error));
        }
    }